/**************************************************************************************************/
/** Style Customization ***************************************************************************/
/**************************************************************************************************/

/* Custom colors */
[data-md-color-scheme="default"] {
    --md-primary-fg-color:        #009682;
    --md-accent-fg-color:         #009682;
    --md-footer-bg-color--dark:   #ffffff;
    --md-footer-fg-color:         #000000;
    --md-footer-fg-color--light:  #00000090;
    --md-footer-fg-color--lighter:#00000050;
    --doc-heading-color:          #DDD;
    --doc-heading-border-color:   #CCC;
    --doc-heading-color-alt:      #F0F0F0;
}
[data-md-color-scheme="slate"] {
    --md-primary-fg-color:        #009682;
    --md-accent-fg-color:         #009682;
    --doc-heading-color:          #000000;
    --doc-heading-border-color:   #000000;
    --doc-heading-color-alt:      #000000;
    --md-code-bg-color:           #000000;
}
[data-md-color-scheme="default"] .md-typeset hr {
    border-bottom-color: #000000;
    border-bottom-width: 1pt;
}
[data-md-color-scheme="slate"] .md-typeset hr {
    border-bottom-color: #ffffff;
}
h4.doc-heading {
    background-color: var(--doc-heading-color);
    border: solid var(--doc-heading-border-color);
    border-width: 1.5pt;
    border-radius: 2pt;
    padding: 0pt 5pt 2pt 5pt;
}
h5.doc-heading, h6.heading {
    background-color: var(--doc-heading-color-alt);
    border-radius: 2pt;
    padding: 0pt 5pt 2pt 5pt;
}

/* Fix view issues */
html {
    scroll-padding-top: 50px;
}

/* Emphasize Headers in navigation bar */
nav.md-nav {
    padding-left: 5px;
}
nav.md-nav--secondary {
    border-left: revert !important;
}
.md-nav__title {
    font-size: 0.9rem;
}
.md-nav__item--section > .md-nav__link {
    font-size: 0.9rem;
}

/* Indent autogenerated documentation */
div.doc-contents {
    padding-left: 25px;
    border-left: 4px solid rgba(230, 230, 230);
}

/* More space at the bottom of the page */
.md-main__inner {
    margin-bottom: 1.5rem;
}
  
/* Remove prev/next footer buttons */
.md-footer__inner {
    display: none;
}

/* Remove superfluous content for docstrings in blocks */
.md-typeset details .mkdocstrings > h4 {
    display: none;
}

.md-typeset details .mkdocstrings > h5 {
    display: none;
}